"""Convert to/from LVC-NR format

The LVC-NR format was first described in https://arxiv.org/abs/1703.01076.  The de
facto format definition is in the code that checks it, which seems to be the
`lvcnrpy` module — more specifically, the classes in lvcnrpy.format.specs at
https://git.ligo.org/waveforms/lvcnrpy

Note that most of the code in this submodule is adapted from code written by Alyssa
Garcia, Geoffrey Lovelace, and Patricia Schmidt.

"""

from .dataset import Dataset
from .waveform_amp_phase import WaveformAmpPhase
from . import waveforms, horizons, metadata, comparisons, conversion
from .comparisons import compare
from .conversion import SimulationConverter, convert_simulation


def bbh_keys_from_simulation_keys(simulation_keys):
    """Extract BBH simulations from a list of all simulations

    Note that this function is maintained here for precise backwards-compatibility.
    More useful functions may be found in `sxs.utilities`.

    """
    return [simulation_key for simulation_key in simulation_keys
            if simulation_key.split(':')[-2] == "BBH"]



#"""Container for metadata of individual simulations"""

import re
import collections


_valid_identifier_pattern = re.compile(r'\W|^(?=\d)')


def _valid_identifier(key):
    return _valid_identifier_pattern.sub('_', key)


_metadata_key_map = {
    # This should be a dictionary of `valid_identifier: metadata_key` pairs for any pair that isn't
    # adequately covered by replacing underscores with dashes.  For example, this would be a working
    # but unnecessary pair:
    #   'simulation_name': 'simulation-name',
}


def _valid_identifier_to_metadata_key(key):
    return _metadata_key_map.get(key, key.replace('_', '-'))


class Metadata(collections.OrderedDict):
    """Interface to metadata

    Note that the constructor is not generally useful from outside this class.  See
    Metadata.from_file, etc., for more useful initialization functions.

    This object is essentially a `collections.OrderedDict`, with a few extra features:
      1) Keys are always forced to be valid python identifiers, whether setting or
         getting.
      2) There are a few extra methods for constructing these objects from json data or
         files or txt files, and outputting them to json or txt files.
      3) If an attribute does not exist on the object, the keys are searched.  This
         allows for tab-completion on the key names.
      4) Properties (methods without parentheses) that attempt to automatically
         determine the resolution and lev number from the 'simulation-name'.

    """

    @classmethod
    def from_file(cls, file_name, ignore_invalid_lines=False, cache_json=True):
        """Read a file into a Metadata object

        Parameters
        ----------
        file_name : str
            The input `file_name` may either end in `.txt`, for an old-style
            metadata.txt file, or in `.json`, for a JSON file.  If neither ending is
            present, the function searches for files with the given `file_name` plus
            both `.txt` and `.json`, and reads the newer one or at least the existing
            one.
        ignore_invalid_lines : bool, optional
            If True, invalid lines will be ignored; otherwise, an error will be raised.
        cache_json : bool, optional
            If True, and the `.json` file does not exist or is older than the `.txt`
            file, a new `.json` file will be created with contents equivalent to the
            `.txt` file.

        Raises
        ------
        SyntaxError : on  `.txt` parse errors when `ignore_invalid_lines` is  False.

        """
        from pathlib import Path

        path = Path(file_name).expanduser().resolve()
        json_path = path.with_suffix(".json")
        txt_path = path.with_suffix(".txt")

        if path.suffix == ".json":
            return cls.from_json_file(json_path)
        elif path.suffix == ".txt":
            return cls.from_txt_file(txt_path, ignore_invalid_lines=ignore_invalid_lines, cache_json=cache_json)
        else:
            json_exists = json_path.exists()
            txt_exists = txt_path.exists()
            if json_exists and not txt_exists:
                return cls.from_json_file(json_path)
            elif txt_exists and not json_exists:
                return cls.from_txt_file(txt_path, ignore_invalid_lines=ignore_invalid_lines, cache_json=cache_json)
            elif json_exists and txt_exists:
                json_time = json_path.stat().st_mtime
                txt_time = txt_path.stat().st_mtime
                if txt_time > json_time:
                    return cls.from_txt_file(txt_path, ignore_invalid_lines=ignore_invalid_lines, cache_json=cache_json)
                else:
                    return cls.from_json_file(json_path)
            else:
                raise ValueError(f"Could not find file named '{file_name}', '{json_path}', or '{txt_path}'")

    load = from_file

    @classmethod
    def from_json_data(cls, json_data):
        """Read metadata from a file-like object (not a file)

        Parameters
        ----------
        json_data : file-like object
            Note that this cannot be just a file name; it must be a file-like object
            (such as an open file handle).  See the `from_json_file` function if you
            just want to pass the path to a file.

        See Also
        --------
        sxs.Metadata.from_json_file : `.json` files
        sxs.Metadata.from_file : reads `.txt` or `.json` files

        """
        import json
        # noinspection PyTypeChecker
        return json.load(json_data, object_pairs_hook=cls)

    @classmethod
    def from_json_file(cls, json_file):
        """Read metadata.json file

        Parameters
        ----------
        json_file : str
            The path to an `metadata.json` file.

        See Also
        --------
        sxs.Metadata.from_file : reads `.txt` or `.json` files

        """
        import json
        from pathlib import Path
        path = Path(json_file).expanduser().resolve().with_suffix(".json")
        with path.open(mode="r") as metadata_file:
            # noinspection PyTypeChecker
            metadata = json.load(metadata_file, object_pairs_hook=cls)
        metadata["metadata_path"] = str(json_file)
        return metadata

    @classmethod
    def from_txt_file(cls, txt_file, ignore_invalid_lines=False, cache_json=True):
        """Read metadata.txt file

        Parameters
        ----------
        txt_file : str
            The path to an old-style `metadata.txt` file.
        ignore_invalid_lines : bool, optional
            If True, invalid lines will be ignored; otherwise, an error will be raised.
        cache_json : bool, optional
            If True, a new `.json` file will be created with contents equivalent to the
            `.txt` file.

        See Also
        --------
        sxs.Metadata.from_file : reads `.txt` or `.json` files

        Notes
        -----
        A standard metadata.txt file is close to being an executable python script that
        just defines a bunch of constants.  The three main problems with the
        metadata.txt format are

          1) variable names contain dashes, which is the subtraction operator in python,
          2) strings are not enclosed in quotes, and
          3) lists are not enclosed in brackets

        It is easy to correct these problems.  In particular, (1) is resolved by
        changing dashes to underscores in the identifiers.  A bug in SpEC
        metadata.txt files -- whereby some comment lines are missing the initial `#` --
        is also fixed.  There are also occasional other problems, like commas missing
        from lists.  All syntax errors as of this writing are fixed in this function.

        Note that this function is not very flexible when it comes to generalizing the
        syntax of the metadata.txt files.  In particular, it assumes that the
        right-hand sides are either numbers or strings (or lists of either numbers or
        strings).  For example, I think I've seen cases where the eccentricity is given
        as something like "<1e-5".  Since python has no "less-than" type, this is
        converted to a string.  But generally, this does seem to work on metadata.txt
        files in the SXS waveform repository.

        """
        # This is considered the safe way to evaluate strings "containing Python values
        # from untrusted sources without the need to parse the values oneself".  We
        # will use it to parse the right-hand side expressions from the metadata.txt
        # file, once we've appropriately modified them, so that python will get to
        # decide what should be a string, float, int, list, etc.
        import warnings
        from ast import literal_eval

        from pathlib import Path
        path = Path(txt_file).expanduser().resolve().with_suffix(".txt")

        assignment_pattern = re.compile(r"""([-A-Za-z0-9]+)\s*=\s*(.*)""")
        string_pattern = re.compile(r"""[A-DF-Za-df-z<>@]""")  # Ignore "e" and "E" because they may appear in numbers
        multi_space_pattern = re.compile(r"""\s+""")
        metadata = cls()

        with path.open("r") as metadata_file:
            for line in metadata_file:
                # Work around bug where some lines of dashes are missing the comment character
                if line.startswith("-"):
                    continue

                # Process each assignment line, skipping comments and unrecognized lines
                match = assignment_pattern.match(line)
                if match:
                    variable, quantity = match.groups()

                    # It was a stupid choice to make variables contain dashes
                    variable = _valid_identifier(variable)

                    # If `quantity` is an empty string, we should just replace it with an empty list
                    if not quantity or quantity == "\n":
                        quantity = "[]"
                    else:
                        q = quantity.strip()
                        if "[unknown]" in q.lower():
                            metadata[variable] = "NaN"
                            continue
                        elif (q.startswith('"') and q.endswith('"')) or (q.startswith("'") and q.endswith("'")):
                            # If the whole thing is quoted, just leave it as is
                            quantity = q
                        elif string_pattern.search(quantity):
                            # If this is a string, strip whitespace from it, split lists and place
                            # brackets around them, and place quotation marks around each element
                            quantities = [q.strip() for q in quantity.split(",")]
                            if "," in quantity:
                                quantity = "['" + "', '".join(quantities) + "']"
                            else:
                                quantity = "'" + quantities[0] + "'"
                        else:
                            # Otherwise, just place brackets around lists of non-strings
                            quantity = quantity.strip()
                            if "," in quantity:
                                quantity = "[" + quantity + "]"
                            elif " " in quantity:
                                quantity = "[" + multi_space_pattern.sub(',', quantity) + "]"

                    # Add this line to the metadata, whether or not it's been modified
                    try:
                        metadata[variable] = literal_eval(quantity)
                    except SyntaxError as e:
                        message = ("\nWhile parsing {0}, transformed input text:\n".format(txt_file)
                                   + "    " + line.rstrip()
                                   + "\ninto quantity\n"
                                   + "    " + variable + " = " + quantity
                                   + "\nParsing this using `ast.literal_eval` resulted in a SyntaxError.\n")
                        if cache_json and ignore_invalid_lines:
                            cache_json = False
                            message += "JSON caching will be turned off for this file until the error is fixed.\n"
                        warnings.warn(message)
                        if not ignore_invalid_lines:
                            raise e

        if cache_json:
            # Skip the text processing next time, and just go straight to json
            metadata.to_json_file(path.with_suffix(".json"))

        metadata["metadata_path"] = str(txt_file)
        return metadata

    def to_json(self, indent=4, separators=(",", ": ")):
        """Export to JSON string"""
        import json
        return json.dumps(self, indent=indent, separators=separators)

    def to_json_file(self, json_file, indent=4, separators=(",", ": ")):
        """Write to JSON file"""
        from pathlib import Path
        path = Path(json_file).expanduser().resolve().with_suffix(".json")
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w") as f:
            f.write(self.to_json(indent=indent, separators=separators))

    save = to_json_file

    def to_txt(self):
        """Export to string like metadata.txt contents"""
        def deformat(value):
            """Basically undo the nice formatting of `from_txt_file`"""
            if isinstance(value, list):
                return ",".join(["{0}".format(item) for item in value])
            else:
                return f"{value}"
        return "\n".join([f"{_valid_identifier_to_metadata_key(key)} = {deformat(self[key])}" for key in self])

    def to_txt_file(self, txt_file):
        """Write to file in metadata.txt format"""
        from pathlib import Path
        path = Path(txt_file).expanduser().resolve().with_suffix(".txt")
        path.mkdir(parents=True, exist_ok=True)
        with path.open("w") as f:
            f.write(self.to_txt() + "\n")

    def __init__(self, *args, **kwargs):
        """Initialize the OrderedDict, converting all keys to valid identifiers

        Note that the constructor is not generally useful from outside this class.  See
        Metadata.from_file, etc., for more useful initialization functions.

        This function intercepts the allowed args and kwargs and converts any keys
        before simply calling the base class's initialization function.

        """
        import collections
        if len(args) > 0:
            args = list(args)
            if isinstance(args[0], collections.abc.Mapping):
                mapping = args[0]
                args[0] = collections.OrderedDict([(_valid_identifier(key), mapping[key]) for key in mapping])
            else:
                iterable = args[0]
                args[0] = [(_valid_identifier(k), v) for k, v in iterable]
        if len(kwargs) > 0:
            for key in list(kwargs):
                kwargs[_valid_identifier(key)] = kwargs.pop(key)
        super(Metadata, self).__init__(*args, **kwargs)

    def add_com_parameters(self, first_only=True, raise_on_errors=False):
        """Add any translation and boost parameters found in all CoM files in this directory

        Adds a new key `com_parameters` to the top level of the metadata dictionary
        containing the `space_translation` and `boost_velocity` parameters for the COM
        correction.

        Parameters
        ----------
        first_only : bool, optional
            If True, add the first set of parameters directly under the top-level key
            `com_parameters`; otherwise, add separate entries under that key for each
            file ending in `_CoM.h5`.
        raise_on_errors : bool, optional
            If False, suppress any exceptions that happen in the core loop of this
            function; otherwise, raise.

        """
        import os.path
        import glob
        import h5py
        import re
        import ast

        path = os.path.dirname(self.get("metadata_path", "."))
        com_parameters = self.get("com_parameters", {})
        for file_name in reversed(sorted(glob.glob(os.path.join(path, "*_CoM.h5")))):
            try:
                with h5py.File(file_name, "r") as f:
                    for g in f:
                        g_parameters = {}
                        if hasattr(f[g], "attrs") and "space_translation" in f[g].attrs:
                            g_parameters["space_translation"] = list(f[g].attrs["space_translation"])
                        if hasattr(f[g], "attrs") and "boost_velocity" in f[g].attrs:
                            g_parameters["boost_velocity"] = list(f[g].attrs["boost_velocity"])
                        if "History.txt" in f[g]:
                            history = f[g]["History.txt"][()]
                            if hasattr(history, "decode"):
                                history = history.decode()
                            for parameter_name in ["boost_velocity", "space_translation"]:
                                if parameter_name not in g_parameters:
                                    pattern = rf'"{parameter_name}": array\((.*?)\)'
                                    matches = re.search(pattern, history)
                                    if matches:
                                        g_parameters[parameter_name] = ast.literal_eval(matches.group(1))
                        if first_only and "space_translation" in g_parameters and "boost_velocity" in g_parameters:
                            self["com_parameters"] = g_parameters
                            return self
                        if g_parameters:
                            com_parameters["{0}/{1}".format(os.path.basename(file_name), g)] = g_parameters
            except Exception:
                if raise_on_errors:
                    raise
        if com_parameters:
            self["com_parameters"] = com_parameters
        return self

    def add_standard_parameters(self, raise_on_errors=False):
        """Add standard parameters that aren't included in the default metadata fields

        New parameters include 'object_types', 'initial_mass_ratio', and
        'reference_mass_ratio'.  If 'reference_dimensionless_spin*' are not present,
        but the parameters necessary to compute them are, they are also added.
        Finally, we also add 'reference_chi_eff', 'reference_chi1_perp', and
        'reference_chi2_perp'.

        """
        import math
        import numpy as np

        def stringify_nan(number):
            if math.isnan(number):
                return "NaN"
            else:
                return number

        def stringify_nans(array):
            return [stringify_nan(number) for number in array]

        if "object1" in self and "object2" in self:
            self["object_types"] = "".join(sorted([self["object1"].upper(), self["object2"].upper()]))
        if "initial_mass1" in self and "initial_mass2" in self:
            try:
                mass_ratio = float(self["initial_mass1"]) / float(self["initial_mass2"])
                self["initial_mass_ratio"] = stringify_nan(mass_ratio)
            except Exception:
                if raise_on_errors:
                    raise
        if "reference_mass1" in self and "reference_mass2" in self:
            try:
                mass_ratio = float(self["reference_mass1"]) / float(self["reference_mass2"])
                self["reference_mass_ratio"] = stringify_nan(mass_ratio)
            except Exception:
                if raise_on_errors:
                    raise
        if "reference_dimensionless_spin1" not in self:
            if "reference_spin1" in self and "reference_mass1" in self:
                try:
                    self["reference_dimensionless_spin1"] = stringify_nans(
                        np.array(self["reference_spin1"]) / self["reference_mass1"]**2
                    )
                except Exception:
                    if raise_on_errors:
                        raise
        if "reference_dimensionless_spin2" not in self:
            if "reference_spin2" in self and "reference_mass2" in self:
                try:
                    self["reference_dimensionless_spin2"] = stringify_nans(
                        np.array(self["reference_spin2"]) / self["reference_mass2"]**2
                    )
                except Exception:
                    if raise_on_errors:
                        raise
        if "initial_dimensionless_spin1" not in self:
            if "initial_spin1" in self and "initial_mass1" in self:
                try:
                    self["initial_dimensionless_spin1"] = stringify_nans(
                        np.array(self["initial_spin1"]) / self["initial_mass1"]**2
                    )
                except Exception:
                    if raise_on_errors:
                        raise
        if "initial_dimensionless_spin2" not in self:
            if "initial_spin2" in self and "initial_mass2" in self:
                try:
                    self["initial_dimensionless_spin2"] = stringify_nans(
                        np.array(self["initial_spin2"]) / self["initial_mass2"]**2
                    )
                except Exception:
                    if raise_on_errors:
                        raise
        if ("reference_mass1" in self and "reference_mass2" in self and "reference_orbital_frequency" in self
                and "reference_dimensionless_spin1" in self and "reference_dimensionless_spin2" in self):
            try:
                m1 = float(self["reference_mass1"])
                m2 = float(self["reference_mass2"])
                chi1 = np.array(self["reference_dimensionless_spin1"], dtype=float)
                chi2 = np.array(self["reference_dimensionless_spin2"], dtype=float)
                L = np.array(self["reference_orbital_frequency"], dtype=float)
                L /= np.linalg.norm(L)
                chi1L = np.dot(chi1, L)
                chi2L = np.dot(chi2, L)
                chi1perp = np.cross(chi1, L)
                chi2perp = np.cross(chi2, L)
                self["reference_chi_eff"] = stringify_nan((m1*chi1L+m2*chi2L)/(m1+m2))
                self["reference_chi1_perp"] = stringify_nan(np.linalg.norm(chi1perp))
                self["reference_chi2_perp"] = stringify_nan(np.linalg.norm(chi2perp))
            except Exception:
                if raise_on_errors:
                    raise
        return self

    def add_extras(self, raise_on_errors=False):
        """Add information to the metadata from other files in its directory"""
        self.add_com_parameters(raise_on_errors=raise_on_errors)
        self.add_standard_parameters(raise_on_errors=raise_on_errors)
        return self

    def reorder_keys(self, order=None):
        """Return a copy of this object with keys reordered

        It is sometimes nice to reorder the keys of the metadata to display the most
        interesting quantities first.  The usual order output by SpEC, for example,
        hides crucial quantities like the masses and spins after lots of uninteresting
        keys like the author list and various bibtex groups.  This function allows the
        keys to be reordered using exact matches and regular expressions.

        """
        import re
        if order is None:
            order = [
                "url",
                "simulation_name",
                "alternative_names",
                "initial_data_type",
                "eos",
                "object_types",
                "number_of_orbits",
                "reference_mass_ratio",
                "reference_chi_eff",
                "reference_chi1_perp",
                "reference_chi2_perp",
                "reference_eccentricity",
                "reference_dimensionless_spin1",
                "reference_dimensionless_spin2",
                "reference_orbital_frequency",
                "reference_mass1",
                "reference_mass2",
                "reference.*",
            ]
        original_keys = list(self)
        new = type(self)()
        for ordered_key in order:
            if ordered_key in original_keys:
                new[ordered_key] = self[ordered_key]
                original_keys.remove(ordered_key)
            else:
                key_pattern = re.compile(ordered_key)
                for key in list(original_keys):  # Iterate over a *copy* of the original_keys list
                    if key_pattern.match(key):
                        new[key] = self[key]
                        original_keys.remove(key)
        for key in original_keys:
            new[key] = self[key]
        return new

    @classmethod
    def fromkeys(cls, iterable):
        iterable = [(_valid_identifier(k), v) for k, v in iterable]
        return super(Metadata, cls).fromkeys(iterable)

    @property
    def resolution(self):
        """Try to determine the resolution from the "simulation-name" field"""
        import os
        simulation_name = self["simulation_name"]
        last_slash_index = simulation_name.rindex(os.sep)
        return simulation_name[last_slash_index+1:]

    @property
    def lev(self):
        """Try to determine an integer "Lev" number from the "simulation-name" field"""
        resolution = self.resolution
        return int(resolution.replace("Lev", ""))

    @property
    def simulation_group(self):
        """Remove any trailing "/LevN" part of the simulation-name"""
        import os
        simulation_name = self["simulation_name"]
        last_slash_index = simulation_name.rindex(os.sep + "Lev")
        if last_slash_index < 1:
            last_slash_index = len(simulation_name)
        return simulation_name[:last_slash_index]

    ###############################################################################
    ###############################################################################
    #                                                                             #
    # The following methods mirror equivalent methods in OrderedDict, but also    #
    # ensure that any keys are converted to valid identifiers first.  This        #
    # enables this object to be used more like the original metadata.txt format.  #
    #                                                                             #
    ###############################################################################
    ###############################################################################

    def __contains__(self, key):
        return super(Metadata, self).__contains__(_valid_identifier(key))

    def __delattr__(self, name):
        super(Metadata, self).__delattr__(_valid_identifier(name))

    def __delitem__(self, key):
        super(Metadata, self).__delitem__(_valid_identifier(key))

    def __getattribute__(self, name):
        """Include keys as attributes

        This allows retrieval of a key like `md["simulation_name"]` as `md.simulation_name`.

        """
        try:
            return super(Metadata, self).__getattribute__(name)
        except AttributeError as e:
            try:
                return self[_valid_identifier(name)]
            except KeyError:
                raise e

    def __dir__(self):
        """Ensure that the keys are included in tab completion"""
        return list(sorted(set(super(Metadata, self).__dir__()))) + list(self.keys())

    def __getitem__(self, key):
        return super(Metadata, self).__getitem__(_valid_identifier(key))

    def __setattr__(self, name, value):
        name = _valid_identifier(name)
        super(Metadata, self).__setattr__(name, value)

    def __setitem__(self, key, value):
        super(Metadata, self).__setitem__(_valid_identifier(key), value)

    def get(self, key, default=None):
        return super(Metadata, self).get(_valid_identifier(key), default)

    def pop(self, key, default=None):
        return super(Metadata, self).pop(_valid_identifier(key), default)

    def setdefault(self, key, default=None):
        return super(Metadata, self).setdefault(_valid_identifier(key), default)

    def update(self, mapping_or_iterable=None, **kwargs):
        if isinstance(mapping_or_iterable, collections.abc.Mapping):
            mapping_or_iterable = collections.OrderedDict(
                [(_valid_identifier(key), mapping_or_iterable[key]) for key in mapping_or_iterable]
            )
        elif isinstance(mapping_or_iterable, collections.abc.Iterable):
            mapping_or_iterable = [(_valid_identifier(k), v) for k, v in mapping_or_iterable]
        return super(Metadata, self).update(mapping_or_iterable)



arxiv_api_url = 'https://export.arxiv.org/api/query'



def request_data(search_query='', id_list='', start=0, max_results=10):
    import requests
    data = {
        'search_query': search_query,
        'id_list': id_list,
        'start': start,
        'max_results': max_results,
    }
    return requests.post(arxiv_api_url, data=data).text


def get_entry_by_arxiv_id(id):
    import feedparser
    response = request_data(id_list=id)
    parsed = feedparser.parse(response)
    return parsed['entries'][0]


def get_all_entries(search_query='', id_list='', start=0, max_results=100):
    import requests
    data = {
        'search_query': search_query,
        'id_list': id_list,
        'start': start,
        'max_results': max_results,
    }
    result = requests.post(arxiv_api_url, data=data)
    entries = result['entries']
    tr = result['opensearch_totalresults']
    si = result['opensearch_startindex']
    ip = result['opensearch_itemsperpage']
    while si+ip < tr:
        data['start'] = si+ip
        result = requests.post(arxiv_api_url, data=data)
        entries += result['entries']
        tr = result['opensearch_totalresults']
        si = result['opensearch_startindex']
        ip = result['opensearch_itemsperpage']
    return entries


def get_journal_reference(entry, string=''):
    if 'journal_ref' in entry:
        return entry['journal_ref']
    if 'arxiv_journal_ref' in entry:
        return entry['arxiv_journal_ref']
    if 'arxiv_doi' in entry:
        try:
            return get_journal_reference_from_doi(entry['arxiv_doi'], string)
        except:
            pass  # Didn't work for some reason; oh well...
    return ''


def get_submission_comment(entry):
    import re
    if 'arxiv_comment' in entry:
        publication_comment_regex = re.compile(
            r"""(?P<comment>[Aa]ccepted (?:for publication )?(?:by|in|to) |[Ss]ubmitted to |[Ii]n press (?:with )?)(?P<publication>[^,;]*)""")
        comment = entry['arxiv_comment']
        search = publication_comment_regex.search(comment)
        if search:
            comment = search['comment']
            publication = search['publication']
            publication_comment = ', ' + comment[0].lower() + comment[1:] + publication
            return publication_comment
    return ''


def get_journal_reference_from_doi(doi, string=''):
    if 'PhysRev' in doi:
        jr = get_journal_reference_from_phys_rev(doi)
    else:
        jr = get_journal_reference_from_ads(doi)
    if string:
        return string.format(jr)
    else:
        return jr


def get_journal_reference_from_phys_rev(doi):
    import re
    import urllib.request
    reference = {}
    pr = re.search('PhysRev([A-Z])', doi)
    fields = re.compile(' (' + '|'.join(['title', 'pub', 'volume', 'issue', 'year', 'pages'])
                        + ') = {(.*)}')
    if pr:
        reference['pub'] = 'Phys. Rev. {0}'.format(pr.groups()[0])
        url = 'https://journals.aps.org/pr{0}/export/{1}'.format(pr.groups()[0].lower(), doi)
        with urllib.request.urlopen(url) as response:
            html = response.read().decode()
        for line in html.split('\n'):
            f = fields.search(line)
            if f:
                field, value = f.groups()[0], f.groups()[1]
                reference[field] = value
    return reference


def get_journal_reference_from_ads(doi):
    import sys
    import difflib
    import ads
    from sxs.references.journal_abbreviations import journal_abbreviation_pairs
    reference = {'title':'', 'pub':'', 'volume':'', 'issue':'', 'year':'', 'page':''}
    try:
        article = ads.SearchQuery(q="doi:{0}".format(doi), fl=list(reference)).next()
        reference['title'] = article.title
        reference['pub'] = article.pub
        reference['volume'] = article.volume
        reference['issue'] = article.issue
        reference['year'] = article.year
        reference['page'] = article.page
    except Exception as e:
        print('Failed to get ADS entry for doi "{0}"'.format(doi), file=sys.stderr)
    closest_match = difflib.get_close_matches(reference['pub'], journal_abbreviation_pairs, n=1)
    if closest_match:
        reference['pub'] = journal_abbreviation_pairs[closest_match[0]]
    return reference



"""Low-level interface to Invenio-backed APIs

"""

from .login import Login
from .deposit import Deposit
from .records import Records

url_standard = 'https://zenodo.org/'
url_sandbox = 'https://sandbox.zenodo.org/'



"""Interface to Zenodo or similar Invenio-backed repositories

"""


from .api import Login, Deposit, Records
from . import catalog, simannex, surrogatemodeling

# See https://github.com/moble/nb-marine-science for other examples using the Zenodo API
# The other python API interface I found is here: https://github.com/moble/zenodo-python

def translate(sxs_identifier, url=False):
    """Query data.black-holes.org to get the current Zenodo equivalent of the given SXS ID

    Parameters
    ----------
    sxs_identifier: string
        The SXS identifier of the simulation (something like SXS:BBH:1234, or
        SXS:BHNS:0001).
    url: bool [defaults to False]
        If True, return the complete URL to the Zenodo record.  Otherwise, only the
        Zenodo identifier of the record (the last part of the URL and DOI) will be
        returned.

    """
    import requests
    r = requests.get('https://data.black-holes.org/waveforms/data/{0}'.format(sxs_identifier), allow_redirects=False)
    if url:
        return r.headers['Location']
    else:
        return r.headers['Location'][len('https://zenodo.org/record/'):]


def records(*args, **kwargs):
    """List all published records

    By default, this function lists all deposits by the current user, logging in by
    default as with the sxs.zenodo.Login class.  Optional parameters allow for
    different searches.

    A list of dicts is returned, each of which contains the "representation" of the
    record, as described in the API documentation:
    http://developers.zenodo.org/#depositions.

    Parameters
    ----------
    json_output: bool [defaults to False]
        If True, this function returns a JSON string; otherwise, it returns a
        python dict.
    sxs: bool [defaults to False]
        If True, this function looks for all records in the 'sxs' community on
        Zenodo.
    all_versions: bool [defaults to False]
        If True, return all versions, not just the most recent version of each
        record.

    Parameters for .api.Login.search
    --------------------------------
    q: string
        Search query, using Elasticsearch query string syntax.  See
        https://help.zenodo.org/guides/search/ for details.  If the above argument
        `sxs` is True, the string ' communities: "sxs" ' is added to the query.
    status: string
        Filter result based on deposit status (either 'draft' or 'published')
    sort: string
        Sort order ('bestmatch' or 'mostrecent').  Prefix with minus to change form
        ascending to descending (e.g., '-mostrecent').
    page: int
        Page number for pagination
    size: int
        Number of results to return per page.  Note that Zenodo (as of this
        writing) seems to place a hard limit of 9999 responses.  Anything more will
        result in an error.

    All remaining parameters are passed to .api.Login.

    """
    import json
    json_output = kwargs.pop('json_output', kwargs.pop('json_output', False))
    sxs = kwargs.pop('sxs', False)
    all_versions = kwargs.pop('all_versions', False)
    q = kwargs.pop('q', '')
    if sxs:
        q += ' communities: "sxs" '
    status = kwargs.pop('status', None)
    sort = kwargs.pop('sort', None)
    page = kwargs.pop('page', None)
    size = kwargs.pop('size', 9999)
    l = Login(*args, **kwargs)
    r = l.search(q, status, sort, page, size, all_versions=all_versions)
    if json_output:
        return json.dumps(r, indent=4, separators=(',', ': '))
    else:
        return r


def related_identifier_formatter(identifier, relation='isSupplementTo', scheme='url'):
    """Convert an identifier to a representation acceptable to Zenodo (INCOMPLETE)

    NOTE: Because of missing documentation, this function is likely incomplete.  In
    particular, the scheme is assumed to be 'url' for everything except simple DOI
    URLs and arXiv links, which are converted.  It is unclear what Zenodo else
    actually converts, and how.

    """
    if 'https://dx.doi.org/' in identifier:
        identifier = identifier.replace('https://dx.doi.org/', '')
        scheme = 'doi'
    elif identifier.lower().startswith('arxiv:'):
        identifier = 'arXiv:' + identifier[6:]
        scheme = 'arxiv'
    elif identifier.startswith('https://arxiv.org/'):
        identifier = 'arXiv:'+identifier.replace('https://arxiv.org/abs/', '').replace('https://arxiv.org/pdf/', '').replace('https://arxiv.org/', '')
        scheme = 'arxiv'
    elif identifier.startswith('http://arxiv.org/'):
        identifier = 'arXiv:'+identifier.replace('http://arxiv.org/abs/', '').replace('http://arxiv.org/pdf/', '').replace('http://arxiv.org/', '')
        scheme = 'arxiv'
    return {
        'identifier': identifier,
        'relation': relation,
        'scheme': scheme,
    }


def upload(directory,
           exclude=[
               'HorizonsDump.h5', 'RedshiftQuantities.h5', 'SpEC.out',
               'rh_FiniteRadii_CodeUnits.h5', 'rPsi4_FiniteRadii_CodeUnits.h5',
               'rhOverM_Asymptotic_GeometricUnits.h5', 'rMPsi4_Asymptotic_GeometricUnits.h5',
           ],
           sandbox=False, access_token_path=None, skip_checksums='if_file_is_older',
           skip_existing=True, deposition_id=None, ignore_deletion=False,
           access_right='closed', license='CC-BY-4.0',
           creators=[], description='', keywords=[], related_identifiers=[],
           error_on_existing=True, publish=True):
    """Publish or edit a Zenodo entry for an SXS simulation

    This is essentially a wrapper around many of the Zenodo API's functions,
    specialized for SXS systems and intended to account for various possible errors
    or special conditions.

    This function should be able to safely handle
      1) new deposits that Zenodo has not seen previously;
      2) drafts that were started previously but failed for some reason, like a
         spurious Zenodo server failure, or some problem in the data that has now
         been fixed; or
      3) systems that have been published on Zenodo previously but have changed in
         some way, so you want to ensure that the local copy and the version on
         Zenodo are in sync.
      4) systems that have been published on Zenodo previously and have not changed
         at all, but you want to verify that the local copy and the version on
         Zenodo are in sync.

    Most of the parameters to this function are simply passed to other functions.
    For more explanation of these parameters, see the relevant function's
    documentation.  Most commonly, the only parameter you really need to pass is
    the first.  You may also wish to pass the last parameter if you want the
    deposit to be published automatically.

    This function returns a Deposit object, which may be used to examine the
    deposit, change it, or publish it if the final parameter is not given as True.

    Parameters only used in this function
    -------------------------------------
    skip_checksums : bool or 'if_file_is_older' [defaults to 'if_file_is_older']
        If False, an MD5 checksum is run for any file that exists on zenodo and
        locally (unless the file sizes are different).  If 'if_file_is_older', the
        files are assumed to match if the local modification time is earlier than
        the creation date of the deposit on zenodo; otherwise, the checksum is run
        (unless the file sizes are different).  If the file sizes or checksums are
        different, the local file is uploaded to zenodo.
    skip_existing : bool [defaults to True]
        If a record with this name exists already, skip this upload.
    error_on_existing : bool [defaults to False]
        If True, and a record already exists for this system, raise an error.
    publish : bool or 'if_pending' [defaults to True]
        If True and the current version on zenodo is not published, publish it.  If
        the input value is the string 'if_pending', it will be changed to True if
        no other changes are made during this run of the function, or to False if
        other changes are made.  This allows for a delay, to ensure that the files
        have stopped changing before the system is actually published.


    Parameters to `.utilities.find_files`
    -------------------------------------
    directory : string
        Absolute or relative path to a directory containing 'common-metadata.txt'
        listing an SXS identifier starting with 'SXS:BBH:', 'SXS:BHNS:', or
        'SXS:NSNS:' and containing at least one 'metadata.txt' file somewhere in
        its file hierarchy.
    exclude : list of strings [defaults to an empty list]

    Parameters to `.api.login.Login`
    --------------------------------
    sandbox : bool [defaults to False]
    access_token_path : string or None [defaults to None]

    Parameters to `.api.deposit.Deposit`
    ------------------------------------
    deposition_id : string, int, or None [defaults to None]
    ignore_deletion : bool [defaults to False]
        If True and this function call does not succeed in publishing the record,
        the returned Deposit object will issue a warning that it has not been
        published when it is deleted (which may happen after the function returns).
   
    Parameters to `.api.deposit.Deposit.update_metadata`
    ----------------------------------------------------
    access_right : string [defaults to 'open']
    license : string [defaults to 'cc-by']
    creators : string [defaults to empty list]
    description : string [defaults to '']
    keywords : string [defaults to empty list]
        Note that the last three parameters, if not passed to this function, will
        be derived automatically from the 'metadata.txt' files found in the SXS
        system directory; they will be the union of the parameters found in each
        file if there are multiple such files.

    """
    import re
    import os
    import datetime
    import pytz
    from ..utilities import sxs_identifier_regex, inspire, md5checksum, find_files, fit_to_console
    from ..metadata import Metadata
    from .creators import known_creators, creators_emails, default_creators
    default_creators = [{'name': 'SXS Collaboration'}]

    directory = os.path.normpath(directory)
    if not os.path.isdir(directory) and os.path.isfile(directory):
        directory = os.path.dirname(directory)
    if not os.path.isdir(directory):
        print('The input directory name "{0}" does not appear to be a directory.'.format(directory))
        raise ValueError(directory)
    if not os.path.isfile(os.path.join(directory, 'common-metadata.txt')):
        print('Could not find common-metadata.txt in this directory')
        raise ValueError(directory)
    sxs_system_re = re.compile(sxs_identifier_regex)
    sxs_system_type = None
    sxs_system_number = None
    with open(os.path.join(directory, 'common-metadata.txt')) as f:
        for line in f.readlines():
            line = line.strip()
            if 'alternative-names' in line:
                m = sxs_system_re.search(line)
                if m:
                    sxs_system_type = m['simulation_type']
                    sxs_system_number = m['sxs_number']
                    break
    if not sxs_system_type or not sxs_system_number:
        raise ValueError("No SXS identifier found in {0}".format(os.path.join(directory, 'common-metadata.txt')))
    sxs_system = 'SXS:{0}:{1}'.format(sxs_system_type, sxs_system_number)
    if sxs_system_type == 'BBH':
        title = 'Binary black-hole simulation {0}'.format(sxs_system)
        default_description = """Simulation of a black-hole binary system evolved by the <a href="{0}">SpEC code</a>."""
    elif sxs_system_type == 'BHNS':
        title = 'Black-hole neutron-star binary simulation {0}'.format(sxs_system)
        default_description = """Simulation of a black-hole neutron-star binary system evolved by the <a href="{0}">SpEC code</a>."""
    elif sxs_system_type == 'NSNS':
        title = 'Binary neutron-star simulation {0}'.format(sxs_system)
        default_description = """Simulation of a neutron-star binary system evolved by the <a href="{0}">SpEC code</a>."""
    else:
        raise ValueError('Did not recognize SXS system type "{0}"'.format(sxs_system_type)
                         + 'in directory "{0}"; should be BBH, BHNS, or NSNS.'.format(directory))
    print("Beginning work on {0}".format(sxs_system))

    # Log in to zenodo
    l = Login(sandbox=sandbox, access_token_path=access_token_path)

    # Get this deposit and the title
    if deposition_id is not None:
        if skip_existing:
            print('Asking to skip existing uploads, but also asked for deposition id {0}.'.format(deposition_id))
            print('These are contradictory.  Raise an exception.')
            raise ValueError(deposition_id)
        d = l.deposit(deposition_id, ignore_deletion=ignore_deletion)
        title = d.title
    else:
        # Check to see if this simulation exists in the list of the user's deposits or in the sxs community
        matching_deposits = l.search(q='title: "{0}"'.format(title))
        if len(matching_deposits) == 1:
            deposition_id = matching_deposits[0]['id']
            print('A deposit with title "{0}"'.format(title))
            if skip_existing:
                print('has already been started with id {0}.'.format(deposition_id))
                if error_on_existing:
                    raise ValueError(title)
                else:
                    return
            print('has already been started with this login.  Opening it for editing.')
            d = l.deposit(deposition_id, ignore_deletion=ignore_deletion)
        elif len(matching_deposits) > 1:
            print('Multiple deposits titled "{0}" have been found.'.format(title))
            raise ValueError(title)
        elif len(matching_deposits) == 0:
            # Check to see if this simulation is already in sxs but not owned by this login
            records = Records.search('title: "{0}"'.format(title))
            records = [r for r in records if r.get('title', '') == title]  # Ensure *exact* match
            communities = [community.get('identifier', '')
                           for representation in records
                           for community in representation.get('metadata', {}).get('communities', {})]
            if 'sxs' in communities:
                print('A record exists on Zenodo with exactly the name "{0}",'.format(title))
                print('but is apparently owned by someone else in the "sxs" community.')
                print('Please contact the owner of that record if you wish to update it.')
                print('Web link: {0}'.format(records[0]['links']['html']))
                raise ValueError(title)
            elif len(communities) > 0:
                print('A record exists on Zenodo with exactly the name "{0}",'.format(title))
                print('but is apparently not owned by you, nor is it in the "sxs" community.')
                print('Web link: {0}'.format(records[0]['links']['html']))
                raise ValueError(title)
            d = l.deposit(deposition_id=None, ignore_deletion=ignore_deletion)
    print('Working on deposit "{0}"'.format(title))

    # Get the time at which this original deposit was created, for possible comparison to file times below
    deposit_created = datetime.datetime.strptime(d.representation['created'], '%Y-%m-%dT%H:%M:%S.%f%z')  # in UTC

    # Convert each metadata.txt file to a metadata.json file sorted with interesting stuff at the
    # top of the file, so it appears prominently on Zenodo's preview without scrolling.  Do this
    # before checking for new files in case these are new or get changed in the process.
    paths_and_names = find_files(directory, exclude=exclude, include_top_directory_in_name=False)
    authors_emails = set()
    point_of_contact_email = ''
    keywords = set(keywords)
    related_identifiers = set(related_identifiers)
    for path,_ in paths_and_names:
        if os.path.basename(path) == 'metadata.txt':
            json_path = os.path.join(os.path.dirname(path), 'metadata.json')
            print('Converting metadata.txt to JSON in {0}'.format(json_path))
            m = Metadata.from_txt_file(path, cache_json=False).add_extras().reorder_keys()
            del m['metadata_path']  # Don't bother recording the local path to the metadata file
            m.to_json_file(json_path)
            authors_emails |= set(m.get('authors_emails', []))
            point_of_contact_email = m.get('point_of_contact_email', point_of_contact_email)
            keywords |= set(m.get('keywords', []))
            bibtex_keys = m.get('bibtex_keys', m.get('simulation_bibtex_keys', []))
            bib_to_id = inspire.map_bibtex_keys_to_identifiers(bibtex_keys)
            related_identifiers |= set(bib_to_id[bib] for bib in bib_to_id)

    # Get list of creators, keywords, and description
    print('Constructing zenodo metadata')
    if not creators:
        creators = d.metadata.get('creators', [])
        if not creators:
            authors_emails = list(authors_emails)
            if not authors_emails:
                print("No creators found in input arguments, on Zenodo, or in any metadata.txt file.")
                if point_of_contact_email in creators_emails:
                    creators.append(creators_emails[point_of_contact_email])
                    print('Using point-of-contact email to add', creators_emails[point_of_contact_email]['name'])
                elif point_of_contact_email:
                    print('Unknown point-of-contact email: {0}'.format(point_of_contact_email))
                creators.extend(default_creators)
                print('Adding default creators')
            else:
                for author_email in authors_emails:
                    name = ' '.join(author_email.split()[:-1])
                    if name in known_creators:
                        creators.append(known_creators[name])
                    else:
                        # We tried our best; let's just get this over with.  Sorry Dr. van Whatever.
                        name_parts = name.split()
                        first_name = ' '.join(name_parts[:-1])
                        last_name = name_parts[-1]
                        if first_name:
                            creators.append({'name': '{0}, {1}'.format(last_name, first_name)})
                        else:
                            creators.append({'name': last_name})
    keywords = sorted(set(keywords) | set(d.metadata.get('keywords', [])))
    if not description:
        description = d.metadata.get('description', '')
        if not description:
            spec_url = "https://www.black-holes.org/code/SpEC.html"
            description = default_description.format(spec_url)
    communities = d.metadata.get('communities', [])
    if 'sxs' not in [c['identifier'] for c in communities]:
        communities.append({'identifier': 'sxs'})
    related_identifiers = [related_identifier_formatter(rid) for rid in sorted(related_identifiers)]
    print('Finished constructing zenodo metadata')

    # Send Zenodo the metadata before messing with files, in case this deposit is interrupted (e.g.,
    # due to long upload times)
    new_metadata = {
        'title': title,
        'upload_type': 'dataset',
        'access_right': access_right,
        'license': license,
        'communities': communities,
        'description': description,
        'keywords': keywords,
        'creators': creators,
    }
    if related_identifiers:
        new_metadata['related_identifiers'] = related_identifiers
    metadata = d.metadata.copy()
    metadata.update(new_metadata)  # Ensure that fields we haven't changed are still present
    unchanged_metadata = (metadata == d.metadata)
    if unchanged_metadata:
        print('No zenodo metadata changed.')
    else:
        try:
            d.edit()
        except:
            print("Failure to unlock deposit is probably acceptable")
        d.update_metadata(metadata)
        print('Uploaded metadata')

    # Get the list of files we'll be uploading and compare to files already in the deposit to see if
    # any have changed.  If so, we need to create a new version.  Otherwise, we can just edit this
    # version.
    zenodo_filenames = d.file_names
    zenodo_name_prefix = ''
    for file_name in zenodo_filenames:
        if file_name.startswith(sxs_system + '/'):
            zenodo_name_prefix = sxs_system + '/'
            break
    local_paths_and_names = find_files(directory, exclude=exclude, include_top_directory_in_name=False)
    if zenodo_name_prefix:
        local_paths_and_names = [[path, zenodo_name_prefix+name] for path, name in local_paths_and_names]
    if len(local_paths_and_names) == 0:
        print('Zenodo requires that there be at least one file.  None found in {0}.'.format(directory))
        raise ValueError('No files found')
    names_to_delete = sorted(set(zenodo_filenames) - set(name for path, name in local_paths_and_names))
    zenodo_file_sizes = d.file_sizes  # formatted as {filename: size_in_bytes}
    zenodo_file_checksums = d.file_checksums  # formatted as {filename: md5checksum}
    print('Comparing sizes and MD5 checksums')
    local_timezone = pytz.timezone("America/Los_Angeles")  # assuming this is the server's timezone
    for path, name in local_paths_and_names.copy():
        if name in zenodo_file_sizes:
            zenodo_filesize = zenodo_file_sizes[name]
            local_filesize = os.path.getsize(path)
            if zenodo_filesize != local_filesize:
                continue  # short-circuit the md5 check because we know they'll be different
            elif skip_checksums == 'if_file_is_older':
                local_mtime = datetime.datetime.fromtimestamp(os.path.getmtime(path))
                if local_timezone.localize(local_mtime) < deposit_created:
                    local_paths_and_names.remove([path, name])
                    continue  # Just assume these files are the same and skip the checksum
            elif skip_checksums:
                local_paths_and_names.remove([path, name])
                continue  # Just assume these files are the same and skip the checksum
        if name in zenodo_file_checksums:
            zenodo_checksum = zenodo_file_checksums[name]
            local_checksum = md5checksum(path)
            if zenodo_checksum == local_checksum:
                local_paths_and_names.remove([path, name])
    name_to_path_map = {name: path for path, name in local_paths_and_names}
    zenodo_names_to_upload_or_replace = [zf for zf in zenodo_filenames if zf in name_to_path_map]
    names_to_upload = sorted(set(name_to_path_map) - set(zenodo_filenames))
    names_to_replace = sorted(set(name_to_path_map) & set(zenodo_filenames))
    paths_and_names_to_upload = [[name_to_path_map[name], name] for name in names_to_upload]
    paths_and_names_to_replace = [[name_to_path_map[name], name] for name in names_to_replace]

    # Now, if needed do the file deletions and/or uploads, and publish
    if not names_to_delete and not names_to_upload and not names_to_replace and unchanged_metadata:
        if publish == 'if_pending':
            publish = True
        print('Nothing will change in this deposit.  Just checking that it is published.')
    elif not names_to_delete and not names_to_upload and not names_to_replace:
        if publish == 'if_pending':
            publish = False
        print('Only metadata will change in this deposit.  Proceeding to publication.')
    else:
        if publish == 'if_pending':
            publish = False
        if d.published:
            # If this deposit already has files, we need to create a new deposit to change the files
            print('Changing files that are already present in a published deposit.  Getting a new version.')
            d = d.get_new_version()
        else:
            # Otherwise this is presumably a new deposit, so we can add files directly with no trouble
            print('Uploading files to an unpublished deposit.')
        # Set the new publication_date
        publication_date = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d")
        metadata = d.metadata.copy()
        metadata.update({'publication_date': publication_date})
        d.update_metadata(metadata, refresh_information=False)
        # Make the changes to the files
        print(fit_to_console(names_to_delete, 'Files to delete: ', subsequent_indent='    '))
        print(fit_to_console(names_to_upload, 'Files to upload: ', subsequent_indent='    '))
        print(fit_to_console(names_to_replace, 'Files to replace: ', subsequent_indent='    '))
        for name in names_to_delete:
            print('    Deleting {0}'.format(name))
            d.delete_file(name, refresh_information=False)
        for path, name in paths_and_names_to_upload:
            print('    Uploading {0}'.format(name))
            d.upload_file(path, name=name, skip_checksum=True, refresh_information=False)
        for path, name in paths_and_names_to_replace:
            print('    Replacing {0}'.format(name))
            d.delete_file(name, refresh_information=False)
            d.upload_file(path, name=name, skip_checksum=True, refresh_information=False)

    # Before returning or checking for published state, make sure our copy of the deposit is up to date with zenodo
    d.refresh_information

    # Publish this version
    if publish:
        if not d.published:
            d.publish()
            print('Finished publishing "{0}" to {1}.'.format(title, d.website))
        else:
            print('Already published "{0}" to {1}.'.format(title, d.website))
    else:
        print('Skipping publication, as requested.')

    return d



import sys
import pathlib
import contextlib
import tempfile
import pytest
import json
import sxs

shortest_metadata = "SXS:BBH:0156v5/Lev5/metadata.json"
shortest_metadata_txt = "SXS:BBH:0156v5/Lev5/metadata.txt"


def test_json_conversion():
    with contextlib.redirect_stdout(None):
        sxs.load(shortest_metadata, download=True, cache=True)
        try:
            sxs.load(shortest_metadata_txt, download=True, cache=True)
        except ValueError:
            pass
    path_json = sxs.utilities.cached_path(shortest_metadata)
    path_txt = sxs.utilities.cached_path(shortest_metadata_txt)
    m = sxs.Metadata.from_txt_file(path_txt, cache_json=False)
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = pathlib.Path(temp_dir)
        temp_path = temp_dir_path / "metadata"
        m.to_json_file(temp_path)
        with path_json.open("r") as f1:
            m1 = json.load(f1)
        with temp_path.with_suffix(".json").open("r") as f2:
            m2 = json.load(f2)
        for key in m2:
            if key.startswith("reference_") or  key.startswith("initial_"):
                assert m1[key] == m2[key]



"""Several convenient functions for interacting with files"""

from contextlib import contextmanager


@contextmanager
def lock_file_manager(file_name, mode='a+'):
    """Create and lock `file_name` as context manager
    
    Use this manager as

    >>> with lock_file_manager('my_file.lock'):
    ...     # do stuff here

    The given file will be created, if needed.  While the code inside the `with`
    block is executing, the file is locked using `fnctl.flock`, preventing any
    other process from accessing it.  As soon as that code finishes executing, the
    file is unlocked and closed.  This is true even if the code raises an
    exception.

    Note that the file descriptor is returned, so it is possible to write to the
    locked file, as in

    >>> with lock_file_manager('my_file.lock') as lf:
    ...     print('Hello, world!', file=lf)
    ...     # do other stuff here

    """
    import fcntl
    with open(file_name, mode) as file_descriptor:
        try:
            fcntl.flock(file_descriptor, fcntl.LOCK_EX)
            yield file_descriptor
        finally:
            fcntl.flock(file_descriptor, fcntl.LOCK_UN)


def find_simulation_directories(root):
    """Search for SpEC-simulation output directories under `root`"""
    import os

    # We can't just do this as a list comprehension because of how os.walk works
    simulation_directories = []

    # Now, recursively walk the directories below `root`
    for dirpath, dirnames, filenames in os.walk(root):
        # We look for files named common-metadata.txt to indicate a simulation //may// be present
        if 'common-metadata.txt' in filenames:
            # Confirm that this is a simulation only if Lev* is present
            if any(s.startswith('Lev') for s in dirnames):
                simulation_directories.append(os.path.join(root, dirpath))
                dirnames[:] = []  # Skip any subdirectories

    return simulation_directories


def find_files(top_directory, exclude=[], include_top_directory_in_name=True):
    """Recursively find all files in `top_directory` and give them relative names

    This function returns a list of pairs.  Each pair gives (first) the full path
    to the file, and (second) its name relative to the parent directory of
    `top_directory`.  These are the parameters needed to upload a file to Zenodo:
    first where to find the file, and second where to put it in the Zenodo deposit.

    Parameters
    ----------
    top_directory : string
        Absolute or relative path to the top directory from which the recursive
        search for files begins.
    exclude : list of strings
        Each string is compiled as a regular expression.  The path to each
        directory and file relative to `top_directory` is searched for a match, and
        if found that item is excluded.  In particular, if a directory matches, no
        files from that directory will be uploaded.
    include_top_directory_in_name : bool [defaults to True]
        If True, the name of the top_directory (relative to its parent) will be
        included in the output names.

    """
    import os
    import re
    paths_and_names = []
    exclude = [re.compile(exclusion) for exclusion in exclude]
    top_directory = os.path.abspath(os.path.expanduser(top_directory))
    parent_directory = os.path.dirname(top_directory)
    for root, dirs, files in os.walk(top_directory, topdown=True):
        dirs.sort(key=str.lower)  # Go in case-insensitive alphabetical order
        files.sort(key=str.lower)  # Go in case-insensitive alphabetical order
        for exclusion in exclude:
            for d in dirs:
                if exclusion.search(os.path.relpath(d, top_directory)):
                    dirs.remove(d)
            for f in files:
                if exclusion.search(os.path.relpath(f, top_directory)):
                    files.remove(f)
        for f in files:
            path = os.path.join(root, f)
            if include_top_directory_in_name:
                name = os.path.relpath(path, parent_directory)
            else:
                name = os.path.relpath(path, top_directory)
            paths_and_names.append([path, name])
    return paths_and_names


def md5checksum(file_name):
    """Compute MD5 checksum on a file, even if it is quite large"""
    from hashlib import md5
    hash_md5 = md5()
    with open(file_name, "rb") as f:
        for chunk in iter(lambda: f.read(32768), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()



"""Functions for syncing the SurrogateModeling data to Zenodo

"""

def sync(annex_dir='./', exclude=None, publish='if_pending',
         sandbox=False, access_token_path=None, skip_checksums='if_file_is_older'):
    """Sync the SurrogateModeling annex to Zenodo

    This function first searches for any changes in the data files; if there are
    any, a new version is created on Zenodo, and the relevant files are uploaded,
    deleted, etc.  It then collects the Zenodo metadata (author list, related
    identifiers, etc.) from zenodo_metadata.json, along with the description to be
    placed on the Zenodo record from zenodo_description.html.  These changes are
    made to the Zenodo record, whether the existing record (if no files changed) or
    the new record (if files changed).

    Parameters
    ----------
    annex_dir : str, optional [defaults to './']
        Absolute or relative path to base directory of SurrogateModeling annex with
        data files.

    """
    import os
    import json
    import datetime
    import pytz
    from ..utilities import md5checksum, find_files, fit_to_console
    from .api import Login

    if exclude is None:
        exclude = []

    l = Login(sandbox=sandbox, access_token_path=access_token_path)
    surrogate_record = l.search(q="title:'Binary black-hole surrogate waveform catalog'", size=1, max_pages=1)[0]
    surrogate_id = surrogate_record['id']
    d = l.deposit(surrogate_id, ignore_deletion=True)

    # Get the time at which this deposit was created, for possible comparison to file times below
    deposit_created = datetime.datetime.strptime(d.representation['created'], '%Y-%m-%dT%H:%M:%S.%f%z')  # in UTC

    # Read the metadata from the top-level file 'zenodo_metadata.json'
    stored_keys = ['creators', 'related_identifiers', 'title']
    metadata = d.metadata.copy()
    with open(os.path.join(annex_dir, 'zenodo_metadata.json'), 'r') as f:
        local_metadata = json.load(f)
    unchanged_metadata = True
    for key in stored_keys:
        if d.metadata.get(key, '') != local_metadata.get(key, ''):
            unchanged_metadata = False
            metadata[key] = local_metadata.get(key, '')

    # Read the description from the top-level file 'zenodo_description.html'
    zenodo_description = d.metadata['description']
    with open(os.path.join(annex_dir, 'zenodo_description.html'), 'r') as f:
        local_description = f.read()
    unchanged_description = (zenodo_description == local_description)
    if not unchanged_description:
        metadata.update({'description': local_description})

    # Make any changes
    if unchanged_description and unchanged_metadata:
        print('No zenodo metadata changed.')
    else:
        try:
            d.edit()
        except Exception:
            print("Failure to unlock deposit is probably acceptable")
        d.update_metadata(metadata)
        print('Uploaded new metadata')

    # Get the list of files we'll be uploading and compare to files already in the deposit to see if
    # any have changed.  If so, we need to create a new version.  Otherwise, we can just edit this
    # version.
    zenodo_filenames = [fn for fn in d.file_names if not fn.endswith('.html')]
    local_paths_and_names = find_files(os.path.join(annex_dir, 'data'), exclude=exclude,
                                       include_top_directory_in_name=False)
    if len(local_paths_and_names) == 0:
        print('Zenodo requires that there be at least one file.  None found in {0}.'.format(annex_dir))
        raise ValueError('No files found')
    names_to_delete = sorted(set(zenodo_filenames) - set(name for path, name in local_paths_and_names))
    zenodo_file_sizes = d.file_sizes  # formatted as {filename: size_in_bytes}
    zenodo_file_checksums = d.file_checksums  # formatted as {filename: md5checksum}
    print('Comparing sizes and MD5 checksums')
    local_timezone = pytz.timezone("America/Los_Angeles")  # assuming this is the server's timezone
    for path, name in local_paths_and_names.copy():
        if name in zenodo_file_sizes:
            zenodo_filesize = zenodo_file_sizes[name]
            local_filesize = os.path.getsize(path)
            if zenodo_filesize != local_filesize:
                continue  # short-circuit the md5 check because we know they'll be different
            elif skip_checksums == 'if_file_is_older':
                local_mtime = datetime.datetime.fromtimestamp(os.path.getmtime(path))
                if local_timezone.localize(local_mtime) < deposit_created:
                    local_paths_and_names.remove([path, name])
                    continue  # Just assume these files are the same and skip the checksum
            elif skip_checksums:
                local_paths_and_names.remove([path, name])
                continue  # Just assume these files are the same and skip the checksum
        if name in zenodo_file_checksums:
            zenodo_checksum = zenodo_file_checksums[name]
            local_checksum = md5checksum(path)
            if zenodo_checksum == local_checksum:
                local_paths_and_names.remove([path, name])
    name_to_path_map = {name: path for path, name in local_paths_and_names}
    names_to_upload = sorted(set(name_to_path_map) - set(zenodo_filenames))
    names_to_replace = sorted(set(name_to_path_map) & set(zenodo_filenames))
    paths_and_names_to_upload = [[name_to_path_map[name], name] for name in names_to_upload]
    paths_and_names_to_replace = [[name_to_path_map[name], name] for name in names_to_replace]

    # Now, if needed do the file deletions and/or uploads, and publish
    if not names_to_delete and not names_to_upload and not names_to_replace and unchanged_description:
        if publish == 'if_pending':
            publish = True
        print('Nothing will change in this deposit.  Just checking that it is published.')
    elif not names_to_delete and not names_to_upload and not names_to_replace:
        if publish == 'if_pending':
            publish = False
        print('Only metadata will change in this deposit.  Proceeding to publication.')
    else:
        if publish == 'if_pending':
            publish = False
        if d.published:
            # If this deposit already has files, we need to create a new deposit to change the files
            print('Changing files that are already present in a published deposit.  Getting a new version.')
            d = d.get_new_version()
        else:
            # Otherwise this is presumably a new deposit, so we can add files directly with no trouble
            print('Uploading files to an unpublished deposit.')
        # Set the new publication_date
        publication_date = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d")
        metadata = d.metadata.copy()
        metadata.update({'publication_date': publication_date})
        d.update_metadata(metadata, refresh_information=False)
        # Make the changes to the files
        print(fit_to_console(names_to_delete, 'Files to delete: ', subsequent_indent='    '))
        print(fit_to_console(names_to_upload, 'Files to upload: ', subsequent_indent='    '))
        print(fit_to_console(names_to_replace, 'Files to replace: ', subsequent_indent='    '))
        for name in names_to_delete:
            print('    Deleting {0}'.format(name))
            d.delete_file(name, refresh_information=False)
        for path, name in paths_and_names_to_upload:
            print('    Uploading {0}'.format(name))
            d.upload_file(path, name=name, skip_checksum=True, refresh_information=False)
        for path, name in paths_and_names_to_replace:
            print('    Replacing {0}'.format(name))
            d.delete_file(name, refresh_information=False)
            d.upload_file(path, name=name, skip_checksum=True, refresh_information=False)

    # Before returning or checking for published state, make sure our copy of the deposit is up to date with zenodo
    # noinspection PyStatementEffect
    d.refresh_information

    # Publish this version
    if publish:
        if not d.published:
            d.publish()
            print('Finished publishing "{0}" to {1}.'.format(d.title, d.website))
        else:
            print('Already published "{0}" to {1}.'.format(d.title, d.website))
    else:
        print('Skipping publication, as requested.')

    return d



import pytest
import sxs


def test_format():
    import pathlib
    import tempfile
    import json
    import h5py
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = pathlib.Path(temp_dir)

        # Test wrong type
        temp_file = type("Failure", (object,), {})
        with pytest.raises(TypeError):
            sxs.file_format(temp_file)

        # Test missing file
        temp_file = temp_path / "nonexistent.json"
        with pytest.raises(FileNotFoundError):
            sxs.file_format(temp_file)

        # Test wrong file format
        temp_file = temp_path / "failure.json"
        with temp_file.open("wb") as f:
            f.write(b"074132708iuhufahc")
        with pytest.raises(ValueError):
            sxs.file_format(temp_file)
        with pytest.raises(ValueError):
            sxs.file_format(str(temp_file))
        with temp_file.open("r") as f:
            with pytest.raises(ValueError):
                sxs.file_format(f)
        with temp_file.open("rb") as f:
            with pytest.raises(ValueError):
                sxs.file_format(f)

        # Test `sxs_format` JSON
        temp_file = temp_path / "test.json"
        version_in = "abcd.efgh"
        with temp_file.open("w") as f:
            json.dump({"sxs_format": version_in}, f)
        version_out = sxs.file_format(temp_file)
        assert version_out == version_in
        version_out = sxs.file_format(str(temp_file))
        assert version_out == version_in
        with temp_file.open("r") as f:
            version_out = sxs.file_format(f)
        assert version_out == version_in

        # Test `sxs_format` JSON
        temp_file = temp_path / "test.json"
        version_in = "efgh.abcd"
        with temp_file.open("w") as f:
            json.dump({"format": version_in}, f)
        version_out = sxs.file_format(temp_file)
        assert version_out == version_in
        version_out = sxs.file_format(str(temp_file))
        assert version_out == version_in
        with temp_file.open("r") as f:
            version_out = sxs.file_format(f)
        assert version_out == version_in

        # Test `sxs_format` HDF5
        temp_file = temp_path / "test.h5"
        version_in = "abef.cdgh"
        with h5py.File(temp_file, "w") as f:
            f.attrs["sxs_format"] = version_in
        version_out = sxs.file_format(temp_file)
        assert version_out == version_in
        version_out = sxs.file_format(str(temp_file))
        assert version_out == version_in
        with temp_file.open("rb") as f:
            version_out = sxs.file_format(f)
        assert version_out == version_in

        # Test `format` HDF5
        temp_file = temp_path / "test.h5"
        version_in = "abgh.efcd"
        with h5py.File(temp_file, "w") as f:
            f.attrs["format"] = version_in
        version_out = sxs.file_format(temp_file)
        assert version_out == version_in
        version_out = sxs.file_format(str(temp_file))
        assert version_out == version_in
        with temp_file.open("rb") as f:
            version_out = sxs.file_format(f)
        assert version_out == version_in



"""Functions to load and save waveforms in RPXMB format"""

from ..utilities import default_shuffle_widths


def save(w, file_name=None, L2norm_fractional_tolerance=1e-10, log_frame=None, shuffle_widths=default_shuffle_widths):
    raise ValueError("""Check
    w.t, w.frame, w.data, w.frame_type, w.spin_weight, w.metadata
    w.to_corotating_frame
    w.boost_velocity, w.space_translation
    """)
    1

    # Make sure that we can understand the file_name and create the directory
    if file_name is None:
        # We'll just be creating a temp directory below, to check
        warnings.warn(
            "\nInput `file_name` is None.  Running in temporary directory.\n"
            "Note that this option is mostly for debugging purposes."
        )
    else:
        h5_path = pathlib.Path(file_name).expanduser().resolve().with_suffix(".h5")
        h5_path.parent.mkdir(parents=True, exist_ok=True)

    shuffle = multishuffle(tuple(shuffle_widths))

    if L2norm_fractional_tolerance == 0.0:
        log_frame = np.log(w.frame).ndarray[:, 1:]
    else:
        # We need this storage anyway, so let's just make a copy and work in-place
        w = w.copy()
        if log_frame is not None:
            log_frame = log_frame.copy()

        # Ensure waveform is in corotating frame
        if w.frame_type == "inertial":
            try:
                initial_time = w.t[0]
                relaxation_time = w.metadata.relaxation_time
                max_norm_time = w.max_norm_time()
                z_alignment_region = ((relaxation_time - initial_time) / (max_norm_time - initial_time), 0.95)
            except Exception:
                z_alignment_region = (0.1, 0.95)
            w, log_frame = w.to_corotating_frame(
                tolerance=1e-10, z_alignment_region=z_alignment_region, truncate_log_frame=True
            )
            log_frame = log_frame.ndarray[:, 1:]
        if w.frame_type != "corotating":
            raise ValueError(
                "Frame type of input waveform must be 'corotating' or 'inertial'; " f"it is {w.frame_type_string}"
            )

        # Convert mode structure to conjugate pairs
        w.convert_to_conjugate_pairs()

        # Set bits below the desired significance level to 0
        w.truncate(tol=L2norm_fractional_tolerance)

        # Compute log(frame)
        if log_frame is None:
            log_frame = np.log(w.frame).ndarray[:, 1:]
            power_of_2 = 2 ** (-np.floor(np.log2(L2norm_fractional_tolerance / 10))).astype("int")
            log_frame = np.round(log_frame * power_of_2) / power_of_2

        # Change -0.0 to 0.0 (~.5% compression)
        w.t += 0.0
        w.data.ndarray += 0.0  # Use ndarray so modes object doesn't override scalar addition
        log_frame.ndarray += 0.0  # Use ndarray so quaternionic doesn't override scalar addition

        # XOR successive instants in time
        t = xor(w.t)
        data = xor(w.data, axis=w.data.time_axis)
        log_frame = xor(log_frame, axis=0)

    # Make sure we have a place to keep all this
    with contextlib.ExitStack() as context:
        if file_name is None:
            temp_dir = context.enter_context(tempfile.TemporaryDirectory())
            h5_path = pathlib.Path(f"{temp_dir}") / "test.h5"
        else:
            print(f'Saving H5 to "{h5_path}"')

        # Write the H5 file
        with h5py.File(h5_path, "w") as f:
            if L2norm_fractional_tolerance != 0.0:
                f.attrs["sxs_format"] = f"{sxs_formats[0]}"
                f.attrs["n_times"] = w.n_times
                f.attrs["ell_min"] = w.ell_min
                f.attrs["ell_max"] = w.ell_max
                f.attrs["shuffle_widths"] = np.array(shuffle_widths, dtype=np.uint8)
                # warnings.warn(f'sxs_format is being set to "{sxs_formats[0]}"')
                data = np.void(
                    bz2.compress(
                        shuffle(t).tobytes()
                        + shuffle(data.flatten("F")).tobytes()
                        + shuffle(log_frame.flatten("F")).tobytes()
                    )
                )
                f.create_dataset("data", data=data)
            else:
                compression_options = {
                    "compression": "gzip",
                    "compression_opts": 9,
                    "shuffle": True,
                }
                f.attrs["sxs_format"] = f"{sxs_formats[0]}"
                f.create_dataset("time", data=w.t.view(np.uint64), chunks=(w.n_times,), **compression_options)
                f.create_dataset("modes", data=w.data.view(np.uint64), chunks=(w.n_times, 1), **compression_options)
                f["modes"].attrs["ell_min"] = w.ell_min
                f["modes"].attrs["ell_max"] = w.ell_max
                f["modes"].attrs["spin_weight"] = w.spin_weight
                if log_frame.size > 1:
                    f.create_dataset(
                        "log_frame", data=log_frame.view(np.uint64), chunks=(w.n_times, 1), **compression_options
                    )

        # Get some numbers for the JSON file
        h5_size = h5_path.stat().st_size
        if file_name is None:
            print(f"Output H5 file size: {h5_size:_} B")
        md5sum = md5checksum(h5_path)

        # Write the corresponding JSON file
        json_path = h5_path.with_suffix(".json")
        json_data = {
            "sxs_format": sxs_formats[0],
            "data_info": {
                "data_type": w.data_type,
                "spin_weight": int(w.spin_weight),
                "ell_min": int(w.ell_min),
                "ell_max": int(w.ell_max),
            },
            "transformations": {
                "truncation": L2norm_fractional_tolerance,
                # see below for 'boost_velocity'
                # see below for 'space_translation'
            },
            "version_info": {
                "python": sys.version,
                "numpy": np.__version__,
                "scipy": scipy.__version__,
                "h5py": h5py.__version__,
                "quaternionic": quaternionic.__version__,
                "spherical": spherical.__version__,
                "sxs": __version__,
                # see below 'spec_version_hist'
            },
            "validation": {
                "h5_file_size": h5_size,
                "n_times": w.n_times,
                "md5sum": md5sum
            },
        }
        if hasattr(w, "boost_velocity"):
            json_data["transformations"]["boost_velocity"] = w.boost_velocity.tolist()
        if hasattr(w, "space_translation"):
            json_data["transformations"]["space_translation"] = w.space_translation.tolist()
        if file_name is not None:
            print(f'Saving JSON to "{json_path}"')
        with json_path.open("w") as f:
            json.dump(json_data, f, indent=2, separators=(",", ": "), ensure_ascii=True)

    return w, log_frame


def load(file_name, ignore_validation=True, check_md5=True, transform_to_inertial=True):
    import warnings
    import pathlib
    import bz2
    import json
    import numpy as np
    import h5py
    import quaternionic
    from ..utilities import md5checksum, xor, multishuffle
    from . import WaveformModes

    sxs_formats = {"rotating_paired_xor_multishuffle_bzip2"}

    def invalid(message):
        if ignore_validation:
            pass
        elif ignore_validation is None:
            warnings.warn(message)
        else:
            raise ValueError(message)

    h5_path = pathlib.Path(file_name).expanduser().resolve().with_suffix(".h5")
    json_path = h5_path.with_suffix(".json")

    # This will be used for validation
    h5_size = h5_path.stat().st_size

    if not json_path.exists():
        invalid(f'\nJSON file "{json_path}" cannot be found, but is expected for this data format.')
        json_data = {}
        data_type = "unknown"
        spin_weight = None
    else:
        with open(json_path) as f:
            json_data = json.load(f)

        data_type = json_data.get("data_info", {}).get("data_type", "unknown")
        spin_weight = json_data.get("data_info", {}).get("spin_weight", None)

        # Make sure this is our format
        sxs_format = json_data.get("sxs_format", "")
        if sxs_format not in sxs_formats:
            invalid(
                f"\nThe `sxs_format` found in JSON file is '{sxs_format}';\n"
                f"it should be one of {sxs_formats}."
            )

        # Make sure the expected H5 file size matches the observed value
        json_h5_file_size = json_data.get("validation", {}).get("h5_file_size", 0)
        if json_h5_file_size != h5_size:
            invalid(
                f"\nMismatch between `validation/h5_file_size` key in JSON file ({json_h5_file_size}) "
                f'and observed file size ({h5_size}) of "{h5_path}".'
            )

        # Make sure the expected H5 file hash matches the observed value
        if check_md5:
            md5sum = md5checksum(h5_path)
            json_md5sum = json_data.get("validation", {}).get("md5sum", "")
            if json_md5sum != md5sum:
                invalid(f"\nMismatch between `validation/md5sum` key in JSON file and observed MD5 checksum.")

    with h5py.File(h5_path, "r") as f:
        # Make sure this is our format
        sxs_format = f.attrs["sxs_format"]
        if sxs_format not in sxs_formats:
            raise ValueError(
                f"\nThe `sxs_format` found in H5 file is '{sxs_format}';"
                f"it should be one of {sxs_formats}."
            )

        # Ensure that the 'validation' keys from the JSON file are the same as in this file
        n_times = f.attrs["n_times"]
        json_n_times = json_data.get("validation", {}).get("n_times", 0)
        if json_n_times != n_times:
            invalid(
                f"\nNumber of time steps in H5 file ({n_times}) "
                f"does not match expected value from JSON ({json_n_times})."
            )

        # Read the raw data
        sizeof_float = 8
        sizeof_complex = 2 * sizeof_float
        ell_min = f.attrs["ell_min"]
        ell_max = f.attrs["ell_max"]
        spin_weight = f.attrs.get("spin_weight", spin_weight)
        data_type = f.attrs.get("data_type", data_type)
        shuffle_widths = tuple(f.attrs["shuffle_widths"])
        unshuffle = multishuffle(shuffle_widths, forward=False)
        n_modes = ell_max * (ell_max + 2) - ell_min ** 2 + 1
        i1 = n_times * sizeof_float
        i2 = i1 + n_times * sizeof_complex * n_modes
        uncompressed_data = bz2.decompress(f["data"][...])
        t = np.frombuffer(uncompressed_data[:i1], dtype=np.uint64)
        data_tmp = np.frombuffer(uncompressed_data[i1:i2], dtype=np.uint64)
        log_frame = np.frombuffer(uncompressed_data[i2:], dtype=np.uint64)

    # Unshuffle the raw data
    t = unshuffle(t)
    data_tmp = unshuffle(data_tmp)
    log_frame = unshuffle(log_frame)

    # Reshape and re-interpret the data
    t = t.view(np.float64)
    data_tmp = data_tmp.reshape((-1, n_times)).T
    log_frame = log_frame.reshape((-1, n_times)).T.copy().view(np.float64)

    # Because of the weirdness of complex types, reshaping, and C/F order, we
    # need to create this with the layout we will eventually want, and then
    # read data into it.
    data = np.empty((n_times, n_modes), dtype=complex).view(np.uint64)

    # Un-XOR the data
    xor(t, reverse=True, preserve_dtype=True, out=t)
    xor(data_tmp, reverse=True, axis=0, out=data)
    xor(log_frame, reverse=True, preserve_dtype=True, axis=0, out=log_frame)

    frame = np.exp(quaternionic.array(np.insert(log_frame, 0, 0.0, axis=1)))

    if spin_weight is None:
        warning = f"Spin weight has not been provided for {h5_path}.  This may result in errors with some functions."
        warnings.warn(warning)

    w = WaveformModes(
        data.view(np.complex128),
        time=t,
        time_axis=0,
        modes_axis=1,
        frame=frame,
        frame_type="corotating",
        data_type=data_type,
        m_is_scaled_out=True,
        r_is_scaled_out=True,
        ell_min=ell_min,
        ell_max=ell_max,
        spin_weight=spin_weight,
    )
    w.convert_from_conjugate_pairs()

    if transform_to_inertial:
        w = w.to_inertial_frame()

    w.json_data = json_data
    w.log_frame = log_frame

    return w



"""File I/O for XMB-format horizons.h5 files"""

from ..utilities import default_shuffle_widths


def save(horizons, file, truncate=lambda x: x, shuffle_widths=default_shuffle_widths):
    """Save `Horizons` object as XMB-format horizons.h5 file

    This function saves the input horizons object in a more compact form than the
    older SpEC-format Horizons.h5 file.  Unlike the older format, only one copy of
    the time data is stored and magnitude datasets are not stored.

    Parameters
    ----------
    horizons : sxs.Horizons
        Horizons object to be written to file.
    file : file-like object, string, or pathlib.Path
        Path to the file on disk or a file-like object (such as an open file
        handle) to be written by h5py.File.
    truncate : callable, optional
        Function that truncates the data to a desired accuracy.  This should set
        bits beyond a desired precision to 0 so that they will compress
        effectively.  By default, no truncation is performed.
    shuffle_widths : array_like, optional
        This must be a series of integers totalling 64 (or the bit width of the
        data in the input `horizons`).  This argument is passed to the multishuffle
        routine.

    See Also
    --------
    sxs.horizons.xor_multishuffle_bzip2.load : load the output file format
    sxs.utilities.xor_multishuffle_bzip2 : compresses data

    """
    import bz2
    import numpy as np
    import h5py
    from ..utilities import xor, multishuffle

    shuffle = multishuffle(tuple(shuffle_widths))

    with h5py.File(file, "w") as f:
        f.attrs["sxs_format"] = "horizons.xor_multishuffle_bzip2"
        f.attrs["shuffle_widths"] = np.array(shuffle_widths, dtype=np.uint8)
        for horizon_name in "ABC":
            horizon = horizons[horizon_name]
            if horizon is not None:
                compressor = bz2.BZ2Compressor()
                # Time
                data = compressor.compress(shuffle(xor(horizon.time)).tobytes())
                # 1-d data sets
                for dat in ["areal_mass", "christodoulou_mass"]:
                    data = data + compressor.compress(
                        shuffle(xor(truncate(horizon[dat].ndarray).view(np.uint64))).tobytes()
                    )
                # 2-d data sets
                for dat in ["coord_center_inertial", "dimensionful_inertial_spin", "chi_inertial"]:
                    data = data + compressor.compress(
                        shuffle(xor(truncate(horizon[dat].ndarray).view(np.uint64).flatten("F"))).tobytes()
                    )
                data = data + compressor.flush()
                d = f.create_dataset(horizon_name, data=np.void(data))
                d.attrs["n_times"] = horizon.time.size


def load(file, ignore_format=False):
    """Load SpEC-format Horizons.h5 file as `Horizons` object

    Parameters
    ----------
    file : file-like object, string, or pathlib.Path
        Path to the file on disk or a file-like object (such as an open file
        handle) to be opened by h5py.File.
    ignore_format : bool, optional
        If True, the format attribute in the input file will be ignored, and this
        function will attempt to load the data regardless.  Note that this will
        return incorrect data or raise an exception if the format is incompatible.
        The default is False, meaning that a ValueError will be raised if the
        format is incorrect.

    Returns
    -------
    horizons : sxs.Horizons
        This is a container for the horizon objects.  See Notes below.

    See also
    --------
    sxs.Horizons : Container object for all of the horizons
    sxs.HorizonQuantities : Container objects for each of the horizons
    sxs.horizons.xor_multishuffle_bzip2.save : save to this file format

    Notes
    -----
    The returned object can be indexed just like the original SpEC-format
    Horizons.h5 file:

        horizons["AhA.dir/CoordCenterInertial.dat"]

    However, the `horizons` object also has a more natural and general interface
    that should be preferred for compatibility with other formats, in which the
    same vector-valued function of time can be accessed as

        horizons.A.coord_center_inertial

    See the documentation of `Horizons` and `HorizonQuantities` for more details.

    """
    import bz2
    import numpy as np
    import h5py
    from .. import Horizons, HorizonQuantities
    from ..utilities import xor, multishuffle

    hqs = {}
    with h5py.File(file, "r") as f:
        if not ignore_format:
            if "sxs_format" not in f.attrs:
                raise ValueError(f"Attribute 'sxs_format' not found in '{file}'")
            sxs_format = f.attrs["sxs_format"]
            if sxs_format != "horizons.xor_multishuffle_bzip2":
                raise ValueError(
                    f"\nAttribute 'sxs_format' found in '{file}' is '{sxs_format}'.\n"
                    f"This function only accepts 'horizons.spec_horizons_h5' formats.\n"
                    f"Use a higher-level `load` function to auto-detect the format."
                )

        shuffle_widths = tuple(f.attrs["shuffle_widths"])
        shuffle = multishuffle(shuffle_widths, forward=False)

        for horizon_name in "ABC":
            if horizon_name in f:
                d = f[horizon_name]
                uncompressed_data = bz2.decompress(d[...])
                n_times = d.attrs["n_times"]
                sizeof_float = 8  # in bytes
                bytes_per_series = n_times * sizeof_float
                hq_kwargs = {}
                i = 0

                # scalar data sets
                for dat in ["time", "areal_mass", "christodoulou_mass"]:
                    hq_kwargs[dat] = xor(shuffle(
                        np.frombuffer(uncompressed_data[i: i + bytes_per_series], dtype=np.uint64)
                    ), reverse=True).view(np.float64)
                    i += bytes_per_series

                # 3-vector data sets
                for dat in ["coord_center_inertial", "dimensionful_inertial_spin", "chi_inertial"]:
                    hq_kwargs[dat] = np.asarray(xor(shuffle(
                        np.frombuffer(uncompressed_data[i: i + 3 * bytes_per_series], dtype=np.uint64)
                    ), reverse=True, preserve_dtype=True).reshape((3, -1)).T.view(np.float64), order="C")
                    i += 3 * bytes_per_series

                horizon = HorizonQuantities(**hq_kwargs)

                hqs[horizon_name] = horizon

    return Horizons(**hqs)



"""Modified greedy algorithm"""

def minimal_grid(x, y, tol=1e-6, error_scale=1.0, y_reference=None):
    """Modified greedy algorithm for building a reduced-order spline

    This function is a modification of the algorithm found in `greedy_spline`.
    Rather than choosing the single worst-approximated point to include in the new
    spline on each iteration (as in the standard greedy algorithm), this function
    chooses a set of points based on peaks in the error function.  To do that, it
    uses the scipy `find_peaks` function.

    Parameters
    ----------
    x : array_like[float]
        An ordered series of sampling points.
    y : array_like[float]
        Function values sampled at points given by `x`.
    tol : float, optional
        This function will iteratively increase the number of sampling points until
        the spline formed from those points is within this value of the input data
        at each input sample.  Defaults to 1e-6.
    error_scale : {float, array_like[float]}, optional
        Multiply the error value by this scale before finding peaks or comparing to
        `tol`.  While this can be achieved by rescaling `tol` when this is a float,
        it is also possible to pass an array to this function so that each
        interpolated value may be targeted with a different tolerance.  In that
        case, the input array must have the same size as `x` and `y`.  The default
        value is 1.0.
    y_reference : {None, array_like[float]}, optional
        Reference `y` values with which to compare.  By default, this is precisely
        `y`; an array similar to `y` can be passed instead, so that `y` will be
        used to construct the splines, but `y_reference` will be used to evaluate
        the errors.  This can be helpful, for example, if we are truncating the
        precision of `y` in the output, but want to ensure that the result is still
        within the given tolerance of the original (pre-truncation) `y`.

    Returns
    -------
    include_sample : array_like[bool]
        Boolean array of the same length as `x` determining whether or not the
        corresponding point should be included in a reduced-order spline.  Note
        that numpy arrays can be indexed with this array, so that
        `x[include_sample]` will give the set of knots and `y[include_sample]` will
        give the corresponding set of data values with which to build the
        reduced-order spline.

    """
    import numpy as np
    from scipy.interpolate import CubicSpline as spline
    from scipy.signal import find_peaks

    deg = 3  # This used to be a parameter, before switching to CubicSpline

    if y_reference is None:
        y_reference = y

    error_scale = np.asarray(error_scale)

    if np.ndim(error_scale) == 0:
        if error_scale != 1.0:
            tol /= error_scale[()]
        def error(ydiff):
            return ydiff
    else:
        def error(ydiff):
            return error_scale * ydiff

    # Implementation notes
    # --------------------
    #
    # The following function is the core of this algorithm.  There are reasonable
    # different choices for the parameters passed to `find_peaks`.  In particular
    # the "prominence" argument seems like it should help in some situations by
    # ensuring that only peaks that actually stand out above their neighbors by
    # `tol` will be added.  However, tests on our waveforms suggest that this only
    # saves ~4% in file size, while slowing things down by ~20%.  I also worry that
    # this could leave off peaks from the final product that should have been
    # included to actually get within the tolerances.  The solution to that would
    # be to start with prominence=tol, and then switch to prominence=None for
    # safety.  Again, however, tests suggest that this has precisely no effect on
    # file sizes (meaning that no more points get added on the second round), but
    # slows down the whole thing by ~5%.  I suspect that the reason this is not
    # very helpful is specific to our data, because so many of the peaks are *very*
    # sharp and very large (mostly coming from junk radiation, and made worse by
    # extrapolation).  In any case, I think the tradeoff is in favor of ignoring
    # the prominence for our use case.
    #
    # Also, the `find_peaks` function has a `height` argument that could select
    # only peaks outside of the desired tolerance (i.e., doing essentially what
    # this code does in the lines following each call to `find_peaks`).  Tests show
    # that using that argument somehow actually increases file size by ~0.2%, while
    # slowing down the code by ~7%.  Evidently, I don't understand precisely what
    # that argument does.  But I am confident that the code below does what we want
    # -- in addition to being faster.

    def next_sample(y, y_greedy, sign):
        if sign[0] == 1:
            errors = error(y - y_greedy)
        else:
            errors = error(y_greedy - y)
        peaks = find_peaks(errors)[0]
        peaks = peaks[np.abs(errors[peaks])>tol]
        sign[0] *= -1
        if not peaks.size:
            peaks = find_peaks(-errors)[0]
            peaks = peaks[np.abs(errors[peaks])>tol]
            sign[0] *= -1
            if not peaks.size:
                return None
        return peaks

    # Create an array describing whether or not to include a given sample in the
    # spline.  Start with all False.
    include_sample = np.zeros(len(x), dtype=bool)

    # Initialize greedy algorithm with evenly spaced indices, as many
    # as needed for the desired spline degree.
    include_sample[np.linspace(0, len(x)-1, num=deg+1, dtype=int)] = True

    # Peak-greed algorithm
    sign = [1]
    for _ in range(len(x)):
        # Compute the spline interpolant on the current set of knots
        s = spline(x[include_sample], y[include_sample])

        # Evaluate this spline on all input data points and find peak errors
        i_next = next_sample(y_reference, s(x), sign)

        # Break out of this loop if `tol` is satisfied
        if i_next is None:
            break

        # Include data point that gives largest interpolation error
        include_sample[i_next] = True

    return include_sample



from .journal_abbreviations import (
    journal_macros, journal_abbreviation_pairs, journal_full_names, journal_abbreviations
)

from .arxiv import (
    request_data, get_entry_by_arxiv_id, get_all_entries, get_journal_reference,
    get_submission_comment, get_journal_reference_from_doi
)

from .fairchild_report import format_paper_list_entry

from .references import limit_author_list
